#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-
#


# Version history:
# 199???   first version by ??? 
# 201907   JLS: 
#     new option to check files in all subfolders
#     for files with duplicate filenames, submit only the most recently modified one
#     pause between connections to the server
#     after a server error, try the operation again
#     if a server does not answer, continue with the next server
#     print a lot of info to the screen, so we can debug errors
#     check for local db type
#     added many comments to the code
# 20190926 JLS: added timestamp info following Volodya's suggestion
# 20191218 JLS: fixed -c (use non-default client.ini) not working
# 20200522 JLS: fixed: don't send files with long filenames (more than 32 chars
#     breaks the db); send B files first     
# 20200529 JLS: implemented Volodya's suggestion: always write a log
# 20201006 JLS: fixed undefined "dirnames" if uploading a single file
# 20201207 JLS: more error logging
# 20201211 JLS: skip files which cause an error in infoFromFilename until we can
#     reproduce the error and code a solution 
# 20210121 JLS: keep the log file until MAX_LOG_SIZE is reached, then start from scratch;
#     compress the tmp zips before submitting them
# 20210510 JLS: only submit files with filenames in ascii
# 20211006 JLS: pass -d daysnewer to submit only the files changed in the last
#     daysnewer days
# 20220211 JLS: also look for files ending in bid.cubic, and bid.dwl
# 20230628 JLS: don't send more than FILETRANSFER_FILES_STEP files every time


# 2do:
#     add more options to the command line (mainly, configure the proxy)
#     check the connection to the server without starting all the submission process
#         (see the "alive.py" code)
#     send to the server some basic info of the client (version, ...)
#     merge verbose's log and printNlog's log


VERSION='20230628' #J please keep this tag updated so we can identify the compiled exes!


'''
refresh.py

  -h,--help       Shows this help
  -i,--input=     File to process. If the given file is a directory, all files within will be processed.
  -s,--servers=   Comma separated list of xmlrpc server nicks (defined in .ini config file)
  -b,--brewerid=  Brewerid of file/s to process
  -c,--config=    config (.ini) file to process
  -r,--recursive  recurse infinitely into all subfolders
  -v,--verbosity  verbose level 

Ej:
  refresh.py -b 033 -s iberonesia -i brewernet\datos
'''


#J ----------------------------------------------------------------------------
#J ----------------------------------------------------------------------------
#J ----------------------------------------------------------------------------
import sys
import xmlrpclib
import zipfile
import os
import tempfile
import md5
import datetime
import re

import log

from optparse import OptionParser
from refreshdb import RefreshDbInfo
from refreshdb import DBMRefreshDb
from refreshdb import CSVRefreshDb
from xmlrpcproxy import ProxiedTransport

import util
from util import IniParser
from util import Options
from util import ServerDef
from util import IniException
from util import ParamException
from util import ArraySplitter
from util import printNlog as pnl

import socket

import time #J time.sleep used to give pauses to the server
import traceback #J to debug errors

global RefreshDbImpl
RefreshDbImpl = CSVRefreshDb  #J default db type is CSV. If you change this to DBM, the
                              #J+"open existing db" code block below MUST be also modified

import zlib #J required for zipfile compression ZIP_DEFLATED

#J ----------------------------------------------------------------------------
#J ----------------------------------------------------------------------------
#J ----------------------------------------------------------------------------
#J Variables to control the communication with the server 
FILELIST_STEP=50 #J while generating the local db from scratch, ask the server for this no. of files
FILETRANSFER_STEP=512*1024 #J max size in bytes (NOT kB) of the zip files sent to server
FILETRANSFER_FILES_STEP=50 #J max number of files inide the zip files sent to server
SECS_BETWEEN_SERVER_COM=2 #J seconds between successful communications with the server
SECS_AFTER_SERVER_ERROR=7 #J after a server error, seconds to wait until the next attempt
MAX_SUB_ATTEMPTS=10 #J number of times a failed submission will be attempted again
MAX_RET_ATTEMPTS=20 #J number of times to retry the data retrieval to generate the local db

MAX_LOG_SIZE=256*1024 #J max size in bytes of the log file saved in the brewer pc. 
                      #J once this size is reached, the log file starts again

#J ----------------------------------------------------------------------------
#J ----------------------------------------------------------------------------
#J ----------------------------------------------------------------------------
# Global flags (configured from the command line or the .ini file)
VERBOSITY_LEVEL = log.INFO
LOGGER = None

#J Initialize global variable for the log filename
LOGFILE = None

#J ----------------------------------------------------------------------------
#J ----------------------------------------------------------------------------
#J ----------------------------------------------------------------------------
def report(msg, loglevel = log.INFO):
        ''' Saca un error por pantalla y por el log.
        '''
        if loglevel >= VERBOSITY_LEVEL: pnl(msg,LOGFILE)

        if LOGGER is not None:
                LOGGER.log(loglevel, msg)


#J ----------------------------------------------------------------------------
class Refresh:
 
        #J --------------------------------------------------
        def __init__(self, brewerid, server_def, fullfilenames):
                '''
                Refresca los ficheros generados por el brewer en el servidor.

                brewerid: brewerid de los ficheros
                server_def : contiene las definiciones del servidor: url y ruta hacia la base de datos de refresh.
                fullfilenames : lista de rutas absolutas de los ficheros a enviar.
                '''

                #J server info
                pnl("    Server: "+server_def.url,LOGFILE)

                if server_def.proxy is None:
                        self.server = xmlrpclib.Server(server_def.url + '/xmlrpc/')
                        #J proxy info
                        pnl("    Proxy:  none",LOGFILE)
                else:
                        self.server = xmlrpclib.Server(server_def.url + '/xmlrpc/', transport = ProxiedTransport(server_def.proxy))
                        #J proxy info
                        pnl("    Proxy:  "+server_def.proxy,LOGFILE)

                #J more options        
                self.brewerid = brewerid
                self.fullfilenames = fullfilenames
                self.dbfile_path = server_def.dbfile

                # Statistics
                self.available = len(fullfilenames)
                self.sent = 0
                self.failed = 0

                self.inDatabase = 0 #J approx number of files in local db
                self.pendingAtStart = 0 #J approx no. of files pending to submit: available - inDB
                self.retrieved=0 #J no. of files retrieved while doing local db generation

                self.retTry=1 #J number of attempts so far to retrieve data when generating local db
                self.subTry=1 #J number of attempts so far to retrieve data when submitting files

                #self.log = True

                self.debug("Starting refresh from", brewerid, "to", server_def.url)
                self.debug("Files to process:", fullfilenames)


        #J --------------------------------------------------
        def debug(self, *args):
                report(' '.join([str(x) for x in args]), log.INFO)


        #J --------------------------------------------------
        def refreshFiles(self):
                '''
                Realiza el procesamiento de los ficheros que contiene fullfilenames.
                Se determina si se deben enviar o no y se envan comprimidos los que s.
                Se actualiza la base de datos refresh con los ficheros enviados correctamente.
                '''


                #
                # Si la refreshdb no existe, la regeneramos a partir de la informacin del servidor.
                #
                if not os.path.exists(self.dbfile_path):
                        pnl('!!! Local db not found, will try to generate it using data from EUBREWNET',LOGFILE)
                        pnl('!!! Local db generation can take a lot of time!',LOGFILE)
                        self.debug("Regenerating refreshdb...")
                        self.regenerateRefreshDb(self.fullfilenames, self.dbfile_path)

                        pnl("*** Local db generated successfuly!",LOGFILE)

                #J this was already done before calling refresh
                #if self.fullfilenames == []:
                #        report('No files to send. Nothing done.', log.WARN)
                #        return

                #J ---------------
                #J we have a local db, add new files to it
                pnl("    Local db: "+self.dbfile_path,LOGFILE)

                #J ---------------------------
                #J open existing db
                global RefreshDbImpl
                try:
                    db = RefreshDbImpl(None, self.dbfile_path)
                    db.open()
                    pnl("    Local db type: CSV",LOGFILE)
                except Exception as err:
                    if "line contains NULL byte" in str(err): # hopefully this is a DBM file...
                        RefreshDbImpl = DBMRefreshDb  #J binary DBM type
                        db = RefreshDbImpl(None, self.dbfile_path)
                        db.open()
                        pnl("    Local db type: DBM",LOGFILE)
                    else:
                        pnl("*** Error opening the local db",LOGFILE)
                        raise Exception(str(err))
                #J ---------------------------

                #J approx number of files in the db at the beginning: the file size of the db 
                #J+divided by the size of each line (between 72 and 74)
                self.inDatabase=os.path.getsize(self.dbfile_path)/74
                pnl("    Approx. number of files in local db: "+str(self.inDatabase),LOGFILE)

                #print '    There are approx. '+str(self.inDatabase)+' files to process'
                self.pendingAtStart=self.available-self.inDatabase

                pnl("    Files in folder(s): %d" % ( self.available ),LOGFILE)

                #J begin the submission
                pnl("*** File submission started",LOGFILE)
                files_to_send = []        ### Lista de RefreshDbInfo de los ficheros a enviar.
                size_to_send = 0
                num_to_send = 0

                #J submit in groups of size_to_send size
                for fullfilename in self.fullfilenames:
                        filename = os.path.basename(fullfilename)

                        db_info = db.read(filename)

                        # 202012 JLS: in some rare cases, infoFromFilename might fail, but I have not been able to
                        # reproduce the error, so I can't propose a fix. AR just asked to skip the file in error
                        try:
                            file_info = self.infoFromFilename(fullfilename)
                        except:
                            pnl("!!! Error getting info from file '"+str(filename)+"', skipped",LOGFILE)
                            continue

                        if not self.needSend(db_info, file_info):
                                continue
                        
                        files_to_send.append(file_info)
                        size_to_send += file_info.size
                        num_to_send += 1

                        if (size_to_send > FILETRANSFER_STEP) or (num_to_send > FILETRANSFER_FILES_STEP):
                                self.debug("Sending files:", [ os.path.basename(info.filename) for info in files_to_send ], "  Total size =", size_to_send)
                                self.sendFilesAndUpdate(db, files_to_send)
                                files_to_send = []
                                size_to_send = 0
                                num_to_send = 0
                
                #J remaining files
                if files_to_send != []:
                        self.debug("Sending files:", [ os.path.basename(info.filename) for info in files_to_send ], "  Total size =", size_to_send)
                        self.sendFilesAndUpdate(db, files_to_send)

                #J all done, close local db
                db.close()

                #J summary of the whole process
                title = "Brewer #%03d -> %s" % (self.brewerid, self.server._ServerProxy__host)

                pnl("*** File submission completed successfuly!",LOGFILE)
                pnl("",LOGFILE) 
                pnl("    "+title,LOGFILE)
                pnl("    "+"-" * len(title),LOGFILE)
                pnl("    Files in folder(s): %d" % ( self.available ),LOGFILE)
                pnl("    Files sent: %d" % ( self.sent ),LOGFILE)
                pnl("    Failed: %d" % ( self.failed ),LOGFILE) #J the server and local db md5sums are different


        #J --------------------------------------------------
        def queryFiles(self):
                #J Not used?
                return self.queryFilesImpl(self.fullfilenames)


        #J --------------------------------------------------
        #J local db generation 
        #J --------------------------------------------------
        def queryFilesImpl(self, fullfilenames):
                '''
                Obtiene la informacin que contiene el servidor sobre los ficheros dados.
                Devuelve un diccionario con clave el nombre de fichero y valor el RefreshDbInfo asociado.
                '''

                #
                # Hacemos un listado escalonado. Demasiados ficheros provocan un timeout.
                #

                result = {}
                splitter = ArraySplitter(fullfilenames, FILELIST_STEP)
                for group in splitter:
                        #print '*** now asking for'
                        #print group

                        # al volver de la funcion, result se ha actualizado.
                        self.queryFilesPart(result, group)

                        #print '*** sleeping after successful retrieval' #J
                        time.sleep(SECS_BETWEEN_SERVER_COM) #J give the server a rest

                        #J write the info we just got, before the server crashes :-O
                        db = RefreshDbImpl(None, self.dbfile_path)
                        db.open()
                        for info in result.itervalues():
                               db.write(info.filename, info)
                        db.close()

                        self.retrieved=self.retrieved+len(group)
                        pnl('    Files pending to check: '+\
                                str(self.available-self.retrieved),LOGFILE)

                return result

                
        #J --------------------------------------------------
        def queryFilesPart(self, infomap, fullfilenames):
                '''
                Obtiene la informacion que tiene el servidor sobre una lista de ficheros.
                Efectua la llamada al servidor y actualiza el diccionario que se pasa como parametro con la informaicion.
                Parametros:
                        infomap       : diccionario a rellenar con la informacion del servidor. Clave: nombre de fichero; Valor: RefreshDbInfo asociado.
                        fullfilenames : lista de rutas completas de los ficheros a consultar.
                '''
                filenames = [ os.path.basename(f) for f in fullfilenames ]
                self.debug("Retrieving info about:", filenames)

                #
                # El resultado es un dict con claves y valores:
                # size : tamao del fichero.
                # md5sum : md5sum del fichero.
                # datetime : fecha hora del archivo
                #
                
                #J ------------------------------------------------------------------
                #J in case of server errors, try MAX_RET_ATTEMPTS times before giving up
                while self.retTry<=MAX_RET_ATTEMPTS:

                    try:
                        # getting info from the server is prone to errors -- refresh may crash here! 
                        maplist = self.server.receive_filelist(self.brewerid, filenames)
                        #print '*** info retrieved from EBN'
                        #print maplist
                        break # success: exit the "while myTry" loop

                    #except socket.gaierror: # doesn't really seem to catch connection timeouts?
                    except Exception as err:

                        if "Connection timed out" in str(err):
                            report("ERROR: Could not connect to server. Exiting...")
                            raise Exception('Connection timed out') #J should be catch in main

                        else:
                            pnl("!!! Db creation error: "+str(err),LOGFILE,"l")

                            if MAX_RET_ATTEMPTS-self.retTry==0:
                                pnl('!!! Max number of retries reached, local db not (completely) generated',LOGFILE)
                                sys.exit(1)

                            pnl('!!! Data retrieval to generate the local db failed, trying again ('+\
                                  str(MAX_RET_ATTEMPTS-self.retTry)+' attempts remaining)',LOGFILE)

                            #print '*** sleeping after server error'
                            time.sleep(SECS_AFTER_SERVER_ERROR)

                            self.retTry=self.retTry+1

                            continue
                #J ------------------------------------------------------------------

                #J "mymap" was originally named "map", but "map" is a reserved keyword nowadays! 
                for (fname, mymap) in maplist.iteritems():
                        if mymap is False:
                                continue
                        #
                        # Construimos el DbRefreshInfo
                        # En python2.5 ya hay un parser para fechas.
                        #
                        dt_parts = mymap['datetime'].split(" ")
                        date_parts = [ int(part) for part in dt_parts[0].split("-") ]
                        time_parts = [ int(part) for part in dt_parts[1].split(":") ]
                        dt = datetime.datetime(date_parts[0], date_parts[1], date_parts[2], time_parts[0], time_parts[1], time_parts[2])
                        remote_info = RefreshDbInfo(mymap['filename'], dt, mymap['size'], mymap['md5sum']) 
                        infomap[mymap['filename']] = remote_info
                        
                return
                

        #J --------------------------------------------------
        def regenerateRefreshDb(self, fullfilenames, dbfile_path):
                '''
                Regenera la base de datos local de refresco.
                Pregunta al servidor por la informacion de los ficheros y crea una nueva base de datos.

                Parametros:
                        fullfilenames: rutas completas de los ficheros a consultar.
                        dbfile_path  : ruta completa a la base de datos a generar.
                '''
                #J get data from the server
                files_info = self.queryFilesImpl(fullfilenames)

                #J write to the db after we have all the data from the server
                #J this is quite dangerous, because if there is an error in one of the data
                #J+retrievals, all the information will be lost -- better write after each
                #J+retrieval -> this code has been moved to RefreshDbImpl
                #db = RefreshDbImpl(None, dbfile_path)
                #db.open()
                #for info in files_info.itervalues():
                #        db.write(info.filename, info)
                #db.close()

        #J --------------------------------------------------
        #J end local db generation 
        #J --------------------------------------------------


        #J --------------------------------------------------
        def infoFromFilename(self, fullfilename):
                '''
                Construye una instancia RefreshDbInfo a partir de la ruta de un fichero.
                NOTA: info.filename es la ruta ABSOLUTA del fichero.
                '''
                size = os.path.getsize(fullfilename)
                dt = datetime.datetime.fromtimestamp(os.path.getmtime(fullfilename))
                md5sum = md5.new(open(fullfilename, 'rb').read()).hexdigest()
                info = RefreshDbInfo(fullfilename, dt, size, md5sum)
                return info
                

        #J --------------------------------------------------
        def needSend(self, db_info, file_info):
                '''
                Determina si un fichero debe ser o no enviado al servidor verificando si ha
                cambiado respecto a la versin existente en el servidor. #J rather, in the local database

                db_info: informacin del fichero en la base de datos refresh. Es una instancia RefreshDbInfo
                file_info: informacin del fichero en el sistema de ficheros. Es una instancia RefreshDbInfo
                '''
                self.debug("Comparing:\n\t", db_info, "\n\t", file_info)
                to_compare = db_info

                if to_compare is None:
                        return True

                # Si los ficheros en contenido son iguales, da igual que la fecha sea diferente.
                #if (to_compare.datetime != file_info.datetime or 
                #                (to_compare.md5sum != file_info.md5sum and to_compare.size != file_info.size) ):
                if        (to_compare.md5sum != file_info.md5sum or to_compare.size != file_info.size) :
                        return True

                return False

 
        #J --------------------------------------------------
        #J file submission
        #J --------------------------------------------------
        def sendFilesAndUpdate(self, db, files_to_send):
                
                '''
                Enva un conjunto de ficheros y actualiza los datos en la refreshdb.

                files_to_send: lista de DBRefreshInfo de los ficheros a enviar.
                '''

                #J send files to the server in a zip. also, get the md5sums from the server
                md5sums = self.sendFiles(files_to_send)

                #J compare the local and server md5sums and update the local db if equal
                self.checkAndUpdateRefreshDb(db, md5sums, files_to_send)
                  
                #J update the stat of submitted files
                self.sent += len(files_to_send)

                #J tell something to the user
                if self.pendingAtStart-self.sent>0:
                    pnl('    Files pending to submit (approx): '+\
                        str(self.pendingAtStart-self.sent),LOGFILE)
                else: #J pendingAtStart is likely wrong bc the local db includes files from other dirs
                    pnl('    Files submitted so far: '+str(self.sent),LOGFILE)


        #J --------------------------------------------------
        def sendFiles(self, files_to_send):
                '''
                files_to_send: lista de DBRefreshInfo de los ficheros a enviar.

                Enva un conjunto de ficheros al servidor.
                Los ficheros se envan dentro de un archivo comprimido.

                NOTA: info.filename debe ser la ruta ABSOLUTA del fichero.
                '''

                fnames = []
                zip_fname = tempfile.mktemp()

                report("Creating tmp file: '%s'" % zip_fname, log.DEBUG)
                pnl("    Creating tmp file: '%s'" % zip_fname,LOGFILE,'l')

                zip_file = zipfile.ZipFile(zip_fname, 'w', compression=zipfile.ZIP_DEFLATED)
                for info in files_to_send:
                        fullfilename = info.filename
                        filename = os.path.basename(fullfilename)
                        fnames.append(filename)
                        report("Adding file '%s' to zipfile." % filename, log.INFO)
                        zip_file.write(fullfilename, filename)

                zip_file.close()

                pnl("    tmp file size in kB = "+str(os.path.getsize(zip_fname)/1024),LOGFILE)

                data = xmlrpclib.Binary(open(zip_fname, "rb").read())

                #J ------------------------------------------------------------------
                #J getting info from the server is prone to errors -- refresh may crash here! 
                while self.subTry<=MAX_SUB_ATTEMPTS:
                    try:
                        result = self.server.receive_files(self.brewerid, data) #J md5sum(s) from the server
                        break # success: exit the "while mySubTry" loop

                    except Exception as err:
                        if "Connection timed out" in str(err):
                            raise Exception('Connection timed out') #J should be catch in main

                        else:
                            if MAX_SUB_ATTEMPTS-self.subTry==0:
                                #J we have done all we could, now just exit
                                pnl('!!! Max number of retries reached, will continue on next execution',LOGFILE)
                                sys.exit(1)

                            #print str(err)
                            #from shutil import copyfile
                            #copyfile(zip_fname, "./failed.zip")
                            pnl("!!! Submission error: "+str(err),LOGFILE)
                            pnl('!!! Submission failed, trying again ('+\
                                    str(MAX_SUB_ATTEMPTS-self.subTry)+' attempts remaining)',LOGFILE)

                            #print '*** sleeping after server error'
                            time.sleep(SECS_AFTER_SERVER_ERROR)

                            self.subTry=self.subTry+1

                            continue
                #J ------------------------------------------------------------------

                #print '*** sleeping after successful submission' #J
                time.sleep(SECS_BETWEEN_SERVER_COM) #J give the server a rest

                report("Deleting tmp file: '%s'" % zip_fname, log.DEBUG)
                pnl("    deleting tmp file: '%s'" % zip_fname,LOGFILE,'l')
                os.unlink(zip_fname)

                return result


        #J --------------------------------------------------
        def checkAndUpdateRefreshDb(self, db, md5sums, files_sent):
                '''
                Comprueba que el fichero fue enviado correctamente, actualizando la refreshdb en tal caso.
                db: Base de datos Refresh.
                md5sums: lista de md5sums recibidos por el servidor.
                files_sent: lista de RefreshDbInfo de los ficheros enviados. 

                Si el md5sum recibido es igual al que se obtuvo del fichero, ste fue enviado correctamente.
                Si ocurriera que el fichero fue modificado tras tomar el info, en la siguiente ejecucin 
                se detectar el cambio y se reenviar.
                '''
                for (md5sum, info) in zip(md5sums, files_sent):
                        if (md5sum == info.md5sum):
                                filename = os.path.basename(info.filename) ### Nos aseguramos que usamos el basename
                                db.write(filename, info)
                                #print "*** File "+filename+" sent and added to the local database" #J
                        else:
                                self.debug("Failed transfer for:", info.filename)
                                self.failed += 1

                #
                # Nos aseguramos de que se escriben los datos.
                #
                db.flush()

        #J --------------------------------------------------
        #J end file submission
        #J --------------------------------------------------


#J ----------------------------------------------------------------------------
def usage(argv):
         print __doc__


#J ----------------------------------------------------------------------------
class OptionBuilder(object):
        '''
        Construye un objeto que contiene las opciones definidas en el fichero de configuracin 
        y en los parmetros pasados al programa por lnea de comandos.

        Las opciones definidas en el archivo de configuracin son sobreescritas por las definidas como parmetro.
        Si alguna opcin es necesaria y no se encuentra especificada, se lanza una ParamException.

        Si alguna opcin especificable solamente mediante el fichero de configuracin no est definida, se lanza
        una IniException.
        '''


        #J --------------------------------------------------
        def __init__(self, ini_filename):
                '''
                ini_filename : Nombre por defecto del fichero .ini
                '''
                
                self.iniParser = IniParser(ini_filename)
                self.options = None

                self.servers_str = self.iniParser.servers_str
                self.brewerid_str = self.iniParser.brewerid_str
                self.input_str = self.iniParser.input_str
                self.rundir_str = self.iniParser.rundir_str
                self.config_str = "config"


        #J --------------------------------------------------
        #
        # Propiedades
        #
        def getServerDefs(self):
                return self.iniParser.server_defs

        server_defs = property(getServerDefs)


        #J --------------------------------------------------
        def getOptions(self):
                '''
                Construye el objeto de opciones, analizando tanto el fichero de configuracin como los parmetros.
                '''
                #
                # Leemos primero los parmetros para poder contemplar el especificar un fichero de configuracin diferente.
                #
                global VERBOSITY_LEVEL 

                (paramOptions, paramArgs) = self.parseArgs()

                if paramOptions.config != None:
                        self.iniParser.ini_path = paramOptions.config

                # J 2019/12: fix config file option (-c) not working
                if paramOptions.config:
                    self.iniParser.ini_filename=paramOptions.config
                

                print "    Reading configuration file: "+self.iniParser.ini_filename

                self.iniParser.parseIni()
                self.options = self.iniParser.options

                VERBOSITY_LEVEL = log.ERROR - 10 * paramOptions.verbosity

                self.options.recursive=paramOptions.recursive #J
                self.options.daysnewer=int(paramOptions.daysnewer) #J

                if paramOptions.brewerid is not None:
                        self.options.brewerid = paramOptions.brewerid
                if paramOptions.servers is not None:
                        self.options.servers = paramOptions.servers
                if paramOptions.input is not None:
                        self.options.input = paramOptions.input

                # J keep track of the config filename
                if paramOptions.config is not None:
                        self.options.config=paramOptions.config
                else:
                        self.options.config="client.ini"

                #
                # Comprobamos las opciones mnimas.
                #
                if self.options.servers is None:
                        raise ParamException("-s must be specified")
                if self.options.brewerid is None:
                        raise ParamException("-b must be specified")
                if self.options.input is None:
                        raise ParamException("-i must be specified")
                if self.options.rundir is None:
                        raise IniException("'working_dir' option must be specified in .ini file")

                util.check_input(self.options.input)
                self.options.brewerid = util.check_brewerid(self.options.brewerid)


        #J --------------------------------------------------
        def parseArgs(self):
                self.parser = OptionParser()
                parser = self.parser

                parser.add_option("-" + self.config_str[0], "--" + self.config_str,\
                    dest=self.config_str,\
                    help="Configuration (.ini) file ('client.ini' by default)", metavar="FILE")

                parser.add_option("-" + self.input_str[0], "--" + self.input_str, dest=self.input_str,\
                    help=\
                    "File or directory to process. If directory, all files within will be processed",\
                    metavar="FILE")

                parser.add_option("-" + self.servers_str[0], "--" + self.servers_str,\
                    dest=self.servers_str, help="Comma separated list of xmlrpc server aliases")

                parser.add_option("-" + self.brewerid_str[0], "--" + self.brewerid_str,\
                    dest=self.brewerid_str, help="Brewer ID of file(s) to process")

                parser.add_option("-v", "--verbose", dest='verbosity', default=0, action="count",\
                    help="Verbosity level. The more '-v', the more verbose is the program")

                parser.add_option("-r", "--recursive", dest="recursive", default=False,\
                    action='store_true', help="Process all subdirectories")

                parser.add_option("-d", "--daysnewer", dest="daysnewer", default=0,\
                    help="Only process files newer than daysnewer days")

                return parser.parse_args()


        #J --------------------------------------------------
        def printHelp(self):
                self.parser.print_help()


#J ----------------------------------------------------------------------------
def applyFilter(brewerid, fullfilenames):
        #
        # Esto podra ser ms eficiente...
        #
        #pattern = re.compile(r".*\.%03d$" % brewerid)
        #J 202202 also look for files ending in bid.cubic, and bid.dwl
        pattern = re.compile(r".*\.%03d$|.*\.%03d\.[Dd][Ww][Ll]|.*\.%03d\.[Cc][Uu][Bb][Ii][Cc]$" %(brewerid,brewerid,brewerid))
        result = []
        for f in fullfilenames:
                if pattern.match(f):
                        result.append(f)
        return result


#J ----------------------------------------------------------------------------
def keepMostRecentDuplicate(fullfilenames):
    # keep only the most recent version of files which have the same name
    #+and are found in more than one dir
    #
    # invoke with a list of fullfilenames generated with e.g.
    # fullfilenames=[os.path.join(dp, f) for dp, dn, fn in os.walk(dirname) \
    #                for f in fn if os.path.isfile(os.path.join(dp, f))]
    #
    # requires os and re
    #
    # 201907 JLS

    # get a list of duplicate filenames
    # based on https://stackoverflow.com/a/18515820, not completely sure how it works :-O
    aux_set=set()
    duplicates=set( os.path.basename(x) for x in fullfilenames if os.path.basename(x) in aux_set\
                    or aux_set.add(os.path.basename(x)) )

    # if duplicates are found, keep only the file with the most recent modification time
    if len(duplicates)>0:

        # tell the user what we are going to do with duplicates
        pnl('!!! Files with the same name have been found in different folders',LOGFILE)
        pnl('!!! See the log file '+LOGFILE.name+' for the list of duplicate files',LOGFILE,'s')
        pnl('!!! Only the most recently modified duplicate will be sent to EUBREWNET',LOGFILE)

        # for each duplicate, remove the versions with older modification times 
        for myDuplicate in duplicates:
            pnl('!!! File with duplicates: '+myDuplicate,LOGFILE,'l')

            # get all versions of each duplicate
            osSep=os.sep.replace('\\','\\\\') # stop conflicts between windows' sep
                                              #+and the reg exp's scape char
            allVersions=[f for f in fullfilenames if re.search(osSep+myDuplicate+'$',f)]

            # find the older versions
            allVersions=sorted(allVersions,reverse=True,key=os.path.getmtime)
            pnl('!!! All versions (most recent first): '+", ".join(allVersions),LOGFILE,'l')

            oldVersions=allVersions[1:]

            # remove the older versions from fullfilenames
            map(fullfilenames.remove,oldVersions)

    # return the (updated, if duplicates found) list of fullfilenames
    return fullfilenames


#J ----------------------------------------------------------------------------
def sort_key_b_first(myFile):
    # use with list.sort or sorted(list,...) as key=sort_key_b_first
    # to place the B files first in the list
    # JLS 20200522
    if os.path.basename(myFile)[0]=="B":
        return 0
    else:
        return 1


#J ----------------------------------------------------------------------------
def main(argv):
        #J fall back log: delete if exists
        if os.path.isfile("refresh_error.txt"):
                os.remove("refresh_error.txt")

        #J main loop
        try:
                #J welcome!
                print '=== Data submission to EUBREWNET started'
                print '    Refresh client version: '+VERSION

                #J parse options
                optionbuilder = OptionBuilder('client.ini')
                optionbuilder.getOptions()

                options = optionbuilder.options
                server_defs = optionbuilder.server_defs

                #J init session log 
                brw_str="%03d" %(options.brewerid)
                logfilename='refresh_log.'+brw_str
                global LOGFILE
                if (os.path.isfile(logfilename)) and (os.path.getsize(logfilename)<MAX_LOG_SIZE):
                    LOGFILE=open(logfilename, 'a')
                else:
                    LOGFILE=open(logfilename, 'w')

                #J basic info to log
                pnl('=== Data submission to EUBREWNET started',LOGFILE,"l")
                pnl('    Refresh client version: '+VERSION,LOGFILE,"l")
                pnl("    Reading configuration file: "+options.config,LOGFILE,"l")
                
                #J brewerid & and date
                now=datetime.datetime.now().strftime('%H:%M:%S %Y-%m-%d')
                pnl("*** Work on Brewer#%03d started at %s" %(options.brewerid, now),LOGFILE)

                #J files to process
                if os.path.isfile(options.input):
                        fullfilenames = [ options.input ]
                        dirnames=False
                else:
                        fullfilenames,dirnames = util.files_in_dir(options.input, options.brewerid,\
                                options.recursive, options.daysnewer)

                #J make sure we found something to process
                if len(fullfilenames)==0:
                    pnl("!!! No files found in input folder(s), please check the input option",LOGFILE)
                    sys.exit(1)

                if dirnames:
                    pnl("    Folders to process: "+dirnames,LOGFILE)

                #J keep only the files of this brewer
                fullfilenames = applyFilter(options.brewerid, fullfilenames)

                #J make sure we found something to process -- after checking for this brewer
                if len(fullfilenames)==0:
                    pnl("!!! No files found for this Brewer, please check the brewerid option",LOGFILE)
                    sys.exit(1)

                #J check for duplicates, keep only the most recent, and warn the user
                fullfilenames=keepMostRecentDuplicate(fullfilenames)

                #J put the B files first
                fullfilenames.sort(key=sort_key_b_first)
              
                #J servers to submit
                servers = options.servers.split(',')

                #J loop over servers
                for server in servers:
                        try:
                                server_def = server_defs[server]
                        except KeyError:
                                raise ParamException("Invalid server: " + server)

                        #J ------------------------------------------------------------------
                        #J try all servers
                        pnl('*** Submitting to '+server,LOGFILE)

                        try: 
                            refresh = Refresh(options.brewerid, server_def, fullfilenames)
                            refresh.refreshFiles()
                        except Exception as err:
                            if "Connection timed out" in str(err):
                                pnl("!!! Connection timed out, please check your network settings",LOGFILE)
                                continue # continue with the next server
                            else:
                                pnl("*** Error '"+str(err)+"' found, follows full traceback info",LOGFILE)
                                pnl(str(traceback.format_exc()),LOGFILE)
                                pnl("*** Please submit the log file "+LOGFILE.name+" to eubrewnet@aemet.es",LOGFILE)
                                sys.exit(2)
                        #J ------------------------------------------------------------------

                #J brewerid & and date
                now=datetime.datetime.now().strftime('%H:%M:%S %Y-%m-%d')
                pnl("*** Work on Brewer#%03d finished at %s" %(options.brewerid, now),LOGFILE)

        except ParamException, e:
                pnl("!!! Error: "+str(e),LOGFILE) #!!!!!!!!!!!!!!!!!!!!!
                optionbuilder.printHelp()
                sys.exit(2)

        except IniException, e:
                pnl("!!! Error in .ini file: "+str(e),LOGFILE) #!!!!!!!!!!!!!!!!
                sys.exit(2)
        

#J ----------------------------------------------------------------------------
if __name__ == "__main__":
        main(sys.argv[1:])


